unit poly;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, numericAnalysis, multidim, mathconfig, linearalgebra;

type

  {
   TPolynom
   A polynom of arbitrary degree.
  }
  TPolynom = class(Analysis.TNumericFunction)
   type
    TInterpolatePoint = record
      x, y : Precision;
      d : Byte;
    end;
    TInterpolateInfo = array of TInterpolatePoint;
   strict private
    _co : Array of Precision;
    function readCo(i: word): Precision; inline;
    procedure writeCo(i: word; AValue: Precision); inline;
   public
    constructor create(_degree:word); reintroduce;
    function degree : word; inline;
    property co[i:word] : Precision read readCo write writeCo; default;
    function calc(const x:Precision) : Precision; override;
    function derivate : Analysis.IFunction; override;
    function integrate(c:Precision) : Analysis.IFunction; override;
    function derivative(const x:Precision) : Precision;
    function toString : String; override;
    class function hermiteInterpolation(const ip:TInterpolateInfo) : TPolynom;
  end;

implementation

{ TPolynom }

constructor TPolynom.create(_degree:word);
begin
  inherited create;
  setLength(_co, _degree+1);
  fillChar(_co[0], (_degree+1)*sizeOf(Precision), #0);
end;

function TPolynom.degree: word;
begin
  result := length(_co)-1;
end;

function TPolynom.readCo(i: word): Precision;
begin
  result := _co[i];
end;

procedure TPolynom.writeCo(i: word; AValue: Precision);
begin
  _co[i] := AValue;
end;

function TPolynom.calc(const x: Precision): Precision;
var
  i : integer;
  x_ : Precision;
begin
  x_ := 1;
  result := co[0];
  for i := 1 to degree do begin
    x_ *= x;
    result += co[i]*x_;
  end;
end;

function TPolynom.derivate: Analysis.IFunction;
var
  r : TPolynom;
  i : integer;
begin
  if (degree = 0) then begin
    r := TPolynom.create(0);
    r.co[0] := 0;
  end else begin
    r := TPolynom.create(degree-1);
    for i := 1 to degree do r.co[i-1] := co[i] * i;
  end;
  result := r;
end;

function TPolynom.integrate(c: Precision): Analysis.IFunction;
var
  r : TPolynom;
  i : integer;
begin
  r := TPolynom.create(degree+1);
  for i := 0 to degree do r.co[i+1] := co[i] / (i+1);
  r.co[0] := c;
  result := r;
end;

function TPolynom.derivative(const x: Precision): Precision;
var
  i : integer;
  x_ : Precision;
begin
  x_ := 1;
  result := 0;
  for i := 1 to degree do begin
    result += i*co[i]*x_;
    x_ *= x;
  end;
end;

function TPolynom.toString: String;
var i : integer;
begin
  result := '';
  for i := length(_co)-1 downto 0 do begin
    if (_co[i] = 0) then continue;
    if (_co[i] < 0) then result += ' - '
    else if (result <> '') then result += ' + ';
    result += floattostr(abs(_co[i]));
    if (i > 0) then result += '*x';
    if (i > 1) then result += '^'+inttostr(i);
  end;
end;

class function TPolynom.hermiteInterpolation(const ip : TInterpolateInfo) : TPolynom;
var
  M : LinAlg.TMatrix;
  a, b : LinAlg.TVector;
  i, j : integer;
  x : Precision;
function fak(v, n : Cardinal) : Cardinal;
begin
  Result := 1;
  for n := n downto 1 do begin
    Result := Result * v;
    if (v > 0) then v := v - 1;
  end;
end;
begin
  result := nil;
  i := length(ip);
  setLength(b, i);
  M.init(i, i, INIT_NONE);
  for i := 0 to length(ip)-1 do begin
    x := 1;
    for j := 0 to length(ip)-1 do begin
      M[i, j] := x * fak(j, ip[i].d);
      if (cardinal(j) >= ip[i].d) then x := x * ip[i].x;
    end;
    b[i] := ip[i].y;
  end;
  try
    LinAlg.gauss(M, a, b);
  finally
    M.clear;
  end;
  result := TPolynom.create(length(a));
  for i := 0 to length(a)-1 do result.co[i] := a[i];
end;


end.

